/**
 * The local storage keys to set on the user's browser
 */
const STATS_KEY = 'nyt-wordle-statistics';

/**
 * Shorthand reference for window.localStorage
 */
const ls = window.localStorage;

/**
 * Attempt to read any local storage statistics for this user
 */
function getExistingLocalData() {
  let statistics = {
    gamesPlayed: 0,
  };

  try {
    const data = JSON.parse(ls.getItem(STATS_KEY));
    if (data && data.gamesPlayed) return data;
  } catch (e) {
    return statistics;
  }
  return statistics;
}

/**
 * Determine whether or not incoming stats should overwrite existing stats
 */
function shouldWrite(incomingStats, force) {
  // If the incoming message doesn't have this property, or it's 0, abort
  if (!incomingStats.gamesPlayed) return false;

  const existingStats = getExistingLocalData();

  if (existingStats.gamesPlayed && !force) {
    console.log(
      'WORDLE DATA TRANSFER: Comparing existing game count with incoming data game count...',
      existingStats.gamesPlayed,
      incomingStats.gamesPlayed,
    );
    return incomingStats.gamesPlayed > existingStats.gamesPlayed;
  } else {
    return true;
  }
}

/**
 * Writes the data to the user's local storage and sets the themes
 */
function transferData(data) {
  if (!data.statistics)
    throw new Error('User local data does not contain statistics. Aborting transfer.');

  if (!shouldWrite(data.statistics, data.force)) {
    console.log('WORDLE DATA TRANSFER: Existing data is fresher. Will not overwrite.');
    return;
  }

  // Write statistics
  ls.setItem(STATS_KEY, JSON.stringify(data.statistics));

  // Handle dark mode
  const darkTheme = data.darkTheme;
  window.themeManager.setDarkTheme(darkTheme);
  // Handle color blind mode
  const colorBlindTheme = !!data.colorBlindTheme;
  window.themeManager.setColorBlindTheme(colorBlindTheme);
}

/**
 * Clean up any data in the address bar
 */
function deleteFromAddressBar() {
  window.history.replaceState({}, document.title, new URL(location.pathname, location.href).href);
}

/**
 * Handles all postMessage events to the window
 */
export function startDataTransfer() {
  if (ls) {
    try {
      const params = new Proxy(new URLSearchParams(window.location.search), {
        get: (searchParams, prop) => searchParams.get(prop),
      });

      if (params.data) {
        const data = JSON.parse(params.data);
        console.log(`WORDLE DATA TRANSFER: Accepted event from origin: ${document.referrer}`, data);
        transferData(data);
      } else {
        console.log('WORDLE DATA TRANSFER: No data received');
      }
    } catch (e) {
      console.warn(`WORDLE DATA TRANSFER: Data may not have been transferred: `, e);
    }

    deleteFromAddressBar();
  }
}
