import {
  evaluateGuess,
  getDifferenceInDays,
  getLetterEvaluations,
  getSolutionForDate,
  validateWord,
  ALPHABET,
  checkGuessHardMode,
  getDayOffsetForDate,
} from './gamelogic.js';
import { initSentry } from './utils/sentry';
import gtm from './gtm';
import { updateStats } from './statistics';
import { clearMoogleState, getGameState, updateGameState } from './gamestate';
import { startRefreshTimer, setRefreshFlag } from './game-refresher';
import { startDataTransfer } from './data-transfer';

export const TOAST_ZINDEX = 1000;
export const PAGE_ZINDEX = 2000;
export const MODAL_ZINDEX = 3000;
// Appears on top of everything
export const SYSTEM_TOAST_ZINDEX = 4000;

// Toggle this to true if we need to stop directing traffic to other games
const disableNavMenu = false;

const gameTemplate = document.createElement('template');
gameTemplate.innerHTML = `
  <style>
  .toaster {
    position: absolute;
    top: 10%;
    left: 50%;
    transform: translate(-50%, 0);
    pointer-events: none;
    width: fit-content;
  }
  #game-toaster {
    z-index: ${TOAST_ZINDEX};
  }
  #system-toaster {
    z-index: ${SYSTEM_TOAST_ZINDEX};
  }

  #game {
    width: 100%;
    max-width: var(--game-max-width);
    margin: 0 auto;
    height: calc(100% - var(--header-height));
    display: flex;
    flex-direction: column;
  }
  header {
    display: flex;
    flex-direction: row;
    align-items: center;
    justify-content: space-between;
    flex-wrap: nowrap;
    padding: 0 16px;
    height: var(--header-height);
    color: var(--color-tone-1);
    border-bottom: 1px solid var(--color-tone-4);
  }
  header .title {
    font-family: 'nyt-karnakcondensed';
    font-weight: 700;
    font-size: 37px;
    line-height: 100%;
    letter-spacing: 0.01em;
    text-align: center;
    left: 0;
    right: 0;
    pointer-events: none;
  }
  .menu-left {
    display: flex;
    margin: 0;
    padding: 0;
    align-items: center;
    width: 70px;
    justify-content: flex-start;
  }
  .menu-right {
    display: flex;
    width: 70px;
    justify-content: flex-end;
  }
  #nav-button {
    padding-top: 2px;
  }

  @media (min-width: 415px) {
    header {
      padding: 0px 16px;
    }
  }

  @media (max-width: 360px) {
    header .title {
      font-size: 22px;
      letter-spacing: 0.1rem;
    }
  }

  #board-container {
    display: flex;
    justify-content: center;
    align-items: center;
    flex-grow: 1;
    overflow: hidden;
  }
  #board {
    display: grid;
    grid-template-rows: repeat(6, 1fr);
    grid-gap: 5px;
    padding:10px;
    box-sizing: border-box;
  }
  button.icon {
    background: none;
    border: none;
    cursor: pointer;
    padding: 0 4px;
  }

  #debug-tools {
    position: absolute;
    bottom: 0;
  }

  </style>
  <game-theme-manager>
  <header>
  <div class="menu-left">
    <button id="nav-button" class="icon" aria-label="Navigation menu. Click for links to other NYT Games and our Privacy Policy." tabindex="-1">
      <nav-icon></nav-icon>
    </button>
    <button id="help-button" class="icon" aria-label="Help" tabindex="-1">
      <game-icon icon="help"></game-icon>
    </button>
  </div>
  <div class="title">
    Wordle
  </div>
  <div class="menu-right">
    <button id="statistics-button" class="icon" aria-label="Statistics" tabindex="-1">
      <game-icon icon="statistics"></game-icon>
    </button>
    <button id="settings-button" class="icon" aria-label="Settings" tabindex="-1">
      <game-icon icon="settings"></game-icon>
    </button>
  </div>
</header>
    <div id="game">
        <div id="board-container">
          <div id="board"></div>
        </div>
        <game-keyboard></game-keyboard>
        <game-nav-modal></game-nav-modal>
        <game-modal></game-modal>
        <game-page></game-page>
        <div class="toaster" id="game-toaster"></div>
        <div class="toaster" id="system-toaster"></div>
    </div>
  </game-theme-manager>
  <div id="debug-tools"></div>
`;

const debugToolsTemplate = document.createElement('template');
debugToolsTemplate.innerHTML = `
<button id="reveal">reveal</button>
<button id="shake">shake</button>
<button id="bounce">bounce</button>
<button id="toast">toast</button>
<button id="modal">modal</button>
`;

// TODO: make these attrs?
const NUM_TILES = 5;
export const NUM_ROWS = 6;
export const IN_PROGRESS = 'IN_PROGRESS';
export const WIN = 'WIN';
const FAIL = 'FAIL';
const WINNING_STRINGS = ['Genius', 'Magnificent', 'Impressive', 'Splendid', 'Great', 'Phew'];
const MAX_BOARD_WIDTH_PX = 350;

export class GameApp extends HTMLElement {
  tileIndex = 0;
  rowIndex = 0;
  solution;
  boardState;
  evaluations;
  canInput = true;
  gameStatus = IN_PROGRESS;
  letterEvaluations = {};
  $board;
  $keyboard;
  $game;
  today;
  lastPlayedTs;
  lastCompletedTs;
  hardMode;
  dayOffset;

  constructor() {
    super();
    this.attachShadow({ mode: 'open' });
    this.today = new Date();
    this.refreshTimer = startRefreshTimer();
    initSentry();
    const gameState = getGameState();
    clearMoogleState();
    this.lastPlayedTs = gameState.lastPlayedTs;
    this.lastCompletedTs = gameState.lastCompletedTs;

    // If dayDifference is 1 or more, load a new game
    // If dayDifference is 0, we've already played today
    if (!this.lastPlayedTs || getDifferenceInDays(new Date(this.lastPlayedTs), this.today) >= 1) {
      // Haven't played before or haven't seen today's solution before
      this.boardState = new Array(NUM_ROWS).fill('');
      this.evaluations = new Array(NUM_ROWS).fill(null);
      // this.solution = getRandomSolution();
      this.solution = getSolutionForDate(this.today);
      this.dayOffset = getDayOffsetForDate(this.today);
      this.lastCompletedTs = gameState.lastCompletedTs;
      this.hardMode = gameState.hardMode;
      this.restoringFromLocalStorage = false;
      // Wipe away the old gameState;
      updateGameState({
        rowIndex: this.rowIndex,
        boardState: this.boardState,
        evaluations: this.evaluations,
        solution: this.solution,
        gameStatus: this.gameStatus,
      });
    } else {
      // Have already played today
      this.boardState = gameState.boardState;
      this.evaluations = gameState.evaluations;
      this.rowIndex = gameState.rowIndex;
      this.solution = gameState.solution;
      this.dayOffset = getDayOffsetForDate(this.today);
      this.letterEvaluations = getLetterEvaluations(this.boardState, this.evaluations);
      this.gameStatus = gameState.gameStatus;
      this.lastCompletedTs = gameState.lastCompletedTs;
      this.hardMode = gameState.hardMode;
      if (this.gameStatus !== IN_PROGRESS) {
        this.canInput = false;
      }
      this.restoringFromLocalStorage = true;
    }

    // console.log({ solution: this.solution });
  }

  evaluateRow() {
    if (this.tileIndex !== NUM_TILES) return;
    if (this.rowIndex >= NUM_ROWS) return;

    const $currentRow = this.$board.querySelectorAll('game-row')[this.rowIndex];
    const letters = this.boardState[this.rowIndex];

    if (!validateWord(letters)) {
      $currentRow.setAttribute('invalid', '');
      this.addToast('Not in word list');
      return;
    }

    if (this.hardMode) {
      const { validGuess, errorMessage } = checkGuessHardMode(
        letters,
        this.boardState[this.rowIndex - 1],
        this.evaluations[this.rowIndex - 1],
      );

      if (!validGuess) {
        $currentRow.setAttribute('invalid', '');
        this.addToast(errorMessage || 'Not valid in hard mode');
        return;
      }
    }

    const evaluation = evaluateGuess(letters, this.solution);
    this.evaluations[this.rowIndex] = evaluation;
    this.letterEvaluations = getLetterEvaluations(this.boardState, this.evaluations);
    $currentRow.evaluation = this.evaluations[this.rowIndex];
    this.rowIndex += 1;
    const isFail = this.rowIndex >= NUM_ROWS;
    const isWin = evaluation.every((letter) => letter === 'correct');
    if (isFail || isWin) {
      const isStreak =
        !!this.lastCompletedTs &&
        getDifferenceInDays(new Date(this.lastCompletedTs), new Date()) === 1;
      // console.log({ isStreak });
      updateStats({ isWin, isStreak, numGuesses: this.rowIndex });
      updateGameState({ lastCompletedTs: Date.now() });
      this.gameStatus = isWin ? WIN : FAIL;

      // Force a game refresh on the next interval
      setRefreshFlag();
    }
    this.tileIndex = 0;
    // Disable inputs until the animation completes
    this.canInput = false;

    updateGameState({
      rowIndex: this.rowIndex,
      boardState: this.boardState,
      evaluations: this.evaluations,
      solution: this.solution,
      gameStatus: this.gameStatus,
      // TODO: should this be this.today?
      lastPlayedTs: Date.now(),
    });
  }

  addLetter(letter) {
    if (this.gameStatus !== IN_PROGRESS) return;
    if (!this.canInput) return;
    if (this.tileIndex >= NUM_TILES) return;
    this.boardState[this.rowIndex] += letter;
    const currentRow = this.$board.querySelectorAll('game-row')[this.rowIndex];
    currentRow.setAttribute('letters', this.boardState[this.rowIndex]);
    this.tileIndex += 1;
  }

  removeLetter() {
    // TODO: Check letter validity here
    if (this.gameStatus !== IN_PROGRESS) return;
    if (!this.canInput) return;
    if (this.tileIndex <= 0) return;
    this.boardState[this.rowIndex] = this.boardState[this.rowIndex].slice(
      0,
      this.boardState[this.rowIndex].length - 1,
    );
    const currentRow = this.$board.querySelectorAll('game-row')[this.rowIndex];
    if (this.boardState[this.rowIndex]) {
      currentRow.setAttribute('letters', this.boardState[this.rowIndex]);
    } else {
      currentRow.removeAttribute('letters');
    }
    currentRow.removeAttribute('invalid');
    this.tileIndex -= 1;
  }

  submitGuess() {
    if (this.gameStatus !== IN_PROGRESS) return;
    if (!this.canInput) return;

    if (this.tileIndex !== NUM_TILES) {
      const $currentRow = this.$board.querySelectorAll('game-row')[this.rowIndex];
      $currentRow.setAttribute('invalid', '');
      this.addToast('Not enough letters');
      return;
    }
    this.evaluateRow();
  }

  addToast(text, duration, system = false) {
    const toast = document.createElement('game-toast');
    toast.setAttribute('text', text);
    if (duration) {
      toast.setAttribute('duration', duration);
    }
    // if (style) {
    //   toast.setAttribute('style', style);
    // }
    if (system) {
      this.shadowRoot.querySelector('#system-toaster').prepend(toast);
    } else {
      this.shadowRoot.querySelector('#game-toaster').prepend(toast);
    }
  }

  // Attempts to fit the board to screen;
  sizeBoard() {
    const $boardContainer = this.shadowRoot.querySelector('#board-container');
    const boardWidth = Math.min(
      Math.floor($boardContainer.clientHeight * (NUM_TILES / NUM_ROWS)),
      MAX_BOARD_WIDTH_PX,
    );
    const tileWidth = Math.floor(boardWidth / NUM_TILES);
    const boardHeight = tileWidth * NUM_ROWS;
    this.$board.style.width = `${boardWidth}px`;
    this.$board.style.height = `${boardHeight}px`;
  }

  showStatsModal() {
    const modal = this.$game.querySelector('game-modal');
    const $gameStats = document.createElement('game-stats');
    if (this.gameStatus === WIN && this.rowIndex <= NUM_ROWS) {
      $gameStats.setAttribute('highlight-guess', this.rowIndex);
    }
    $gameStats.gameApp = this;
    modal.appendChild($gameStats);
    modal.setAttribute('open', '');
  }

  showNavModal() {
    const modal = this.$game.querySelector('game-nav-modal');
    const $gameNav = document.createElement('game-nav');
    $gameNav.gameApp = this;
    modal.appendChild($gameNav);
    modal.setAttribute('open', '');
  }

  showHelpModal() {
    const modal = this.$game.querySelector('game-modal');
    modal.appendChild(document.createElement('game-help'));
    modal.setAttribute('open', '');
  }

  connectedCallback() {
    this.shadowRoot.appendChild(gameTemplate.content.cloneNode(true));
    this.$game = this.shadowRoot.querySelector('#game');
    this.$board = this.shadowRoot.querySelector('#board');
    this.$keyboard = this.shadowRoot.querySelector('game-keyboard');

    if (disableNavMenu) {
      this.shadowRoot.querySelector('#nav-button').style.display = 'none';
    }

    this.sizeBoard();

    if (!this.lastPlayedTs) {
      // never played before
      setTimeout(() => this.showHelpModal(), 100);
    }

    for (let i = 0; i < NUM_ROWS; i++) {
      console.groupCollapsed(`${this.boardState[i]} create row`);
      const row = document.createElement('game-row');
      row.setAttribute('letters', this.boardState[i]);
      row.setAttribute('length', NUM_TILES);
      if (this.evaluations[i]) {
        row.evaluation = this.evaluations[i];
      }
      this.$board.appendChild(row);
      console.groupEnd();
    }

    this.$game.addEventListener('game-key-press', (e) => {
      const { key } = e.detail;
      if (key === '←' || key === 'Backspace') {
        this.removeLetter();
      } else if (key === '↵' || key === 'Enter') {
        this.submitGuess();
      } else if (ALPHABET.includes(key.toLowerCase())) {
        this.addLetter(key.toLowerCase());
      }
    });

    this.$game.addEventListener('game-last-tile-revealed-in-row', (e) => {
      this.$keyboard.letterEvaluations = this.letterEvaluations;
      if (this.rowIndex < NUM_ROWS) {
        this.canInput = true;
      }

      // Shows the modal on win, or after reload with a win in localStorage
      const lastRow = this.$board.querySelectorAll('game-row')[this.rowIndex - 1];
      // Only chrome can read event.path
      const eventPath = e.path || (e.composedPath && e.composedPath());

      if (eventPath.includes(lastRow)) {
        if ([WIN, FAIL].includes(this.gameStatus)) {
          // If we're not loading from localStorage, we want to display winning/losing states
          if (!this.restoringFromLocalStorage) {
            if (this.gameStatus === WIN) {
              lastRow.setAttribute('win', '');
              this.addToast(WINNING_STRINGS[this.rowIndex - 1], 2000);
            }
            if (this.gameStatus === FAIL) {
              this.addToast(this.solution.toUpperCase(), Infinity);
            }
            setTimeout(() => {
              this.showStatsModal();
            }, 2500);
          } else {
            this.showStatsModal();
          }
        }
        // once the last row is loaded, we're no longer restoring from local storage
        this.restoringFromLocalStorage = false;
      }
    });

    this.shadowRoot.addEventListener('game-setting-change', (e) => {
      const { name, checked, disabled } = e.detail;
      switch (name) {
        case 'hard-mode': {
          if (disabled) {
            this.addToast('Hard mode can only be enabled at the start of a round', 1500, true);
          } else {
            this.hardMode = checked;
            updateGameState({ hardMode: checked });
          }
          return;
        }
      }
    });

    this.shadowRoot.getElementById('settings-button').addEventListener('click', (e) => {
      const page = this.$game.querySelector('game-page');
      const title = document.createTextNode('Settings');
      page.appendChild(title);
      const content = document.createElement('game-settings');
      content.setAttribute('slot', 'content');
      content.gameApp = this;
      page.appendChild(content);
      page.setAttribute('open', '');
    });

    this.shadowRoot.getElementById('help-button').addEventListener('click', (e) => {
      const page = this.$game.querySelector('game-page');
      const title = document.createTextNode('How to play');
      page.appendChild(title);
      const content = document.createElement('game-help');
      content.setAttribute('page', '');
      content.setAttribute('slot', 'content');
      page.appendChild(content);
      page.setAttribute('open', '');
    });

    this.shadowRoot.getElementById('statistics-button').addEventListener('click', (e) => {
      this.showStatsModal();
    });

    this.shadowRoot.getElementById('nav-button').addEventListener('click', (e) => {
      this.showNavModal();
    });

    window.addEventListener('resize', this.sizeBoard.bind(this));

    // this.debugTools();

    // Kick off data transfer
    startDataTransfer();

    // Kick off gtm script
    gtm();
  }

  disconnectedCallback() {
    // TODO: cleanup event listeners
    clearInterval(this.refreshTimer);
  }

  debugTools() {
    this.shadowRoot
      .getElementById('debug-tools')
      .appendChild(debugToolsTemplate.content.cloneNode(true));

    this.shadowRoot.getElementById('toast').addEventListener('click', (e) => {
      this.addToast('hello world');
    });
    this.shadowRoot.getElementById('modal').addEventListener('click', (e) => {
      const modal = this.$game.querySelector('game-modal');
      modal.textContent = 'hello plz';
      modal.setAttribute('open', '');
    });
    this.shadowRoot.getElementById('reveal').addEventListener('click', () => {
      this.evaluateRow();
    });
    this.shadowRoot.getElementById('shake').addEventListener('click', () => {
      const currentRow = this.$board.querySelectorAll('game-row')[this.rowIndex];
      currentRow.setAttribute('invalid', '');
    });
    this.shadowRoot.getElementById('bounce').addEventListener('click', () => {
      const currentRow = this.$board.querySelectorAll('game-row')[this.rowIndex - 1];
      const attr = currentRow.getAttribute('win');
      if (attr === '') {
        currentRow.removeAttribute('win');
      } else {
        currentRow.setAttribute('win', '');
      }
    });
  }
}

customElements.define('game-app', GameApp);
