import { ALPHABET } from './gamelogic.js'

const keyboardTemplate = document.createElement('template')

keyboardTemplate.innerHTML = `
  <style>
  :host {
    height: var(--keyboard-height);
  }
  #keyboard {
    margin: 0 8px;
    user-select: none;
  }
  
  .row {
    display: flex;
    width: 100%;
    margin: 0 auto 8px;
    /* https://stackoverflow.com/questions/46167604/ios-html-disable-double-tap-to-zoom */
    touch-action: manipulation;
  }
  
  button {
    font-family: inherit;
    font-weight: bold;
    border: 0;
    padding: 0;
    margin: 0 6px 0 0;
    height: 58px;
    border-radius: 4px;
    cursor: pointer;
    user-select: none;
    background-color: var(--key-bg);
    color: var(--key-text-color);
    flex: 1;
    display: flex;
    justify-content: center;
    align-items: center;
    text-transform: uppercase;
    -webkit-tap-highlight-color: rgba(0,0,0,0.3);
  }

  button:focus {
    outline: none;
  }

  button.fade {
    transition: background-color 0.1s ease, color 0.1s ease;
  }
  
  button:last-of-type {
    margin: 0;
  }
  
  .half {
    flex: 0.5;
  }
  
  .one {
    flex: 1;
  }

  .one-and-a-half {
    flex: 1.5;
    font-size: 12px;
  }
  
  .two {
    flex: 2;
  }

  button[data-state='correct'] {
    background-color: var(--key-bg-correct);
    color: var(--key-evaluated-text-color);
  }

  button[data-state='present'] {
    background-color: var(--key-bg-present);
    color: var(--key-evaluated-text-color);
  }

  button[data-state='absent'] {
    background-color: var(--key-bg-absent);
    color: var(--key-evaluated-text-color);
  }

  </style>
  <div id="keyboard"></div>
`

const keyTemplate = document.createElement('template')
keyTemplate.innerHTML = `
  <button>key</button>
`

const spacerTemplate = document.createElement('template')
spacerTemplate.innerHTML = `
  <div class="spacer"></div>
`

const keyboardRows = [
  ['q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p'],
  ['-', 'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', '-'],
  ['↵', 'z', 'x', 'c', 'v', 'b', 'n', 'm', '←'],
]

export class GameKeyboard extends HTMLElement {
  _letterEvaluations = {}

  constructor() {
    super()
    this.attachShadow({ mode: 'open' })
  }

  set letterEvaluations(value) {
    this._letterEvaluations = value
    this._render()
  }

  dispatchKeyPressEvent(key) {
    this.dispatchEvent(
      new CustomEvent('game-key-press', {
        bubbles: true,
        composed: true,
        detail: { key },
      }),
    )
  }

  connectedCallback() {
    this.shadowRoot.appendChild(
      keyboardTemplate.content.cloneNode(true),
    )
    this.$keyboard = this.shadowRoot.getElementById('keyboard')

    this.$keyboard.addEventListener('click', (e) => {
      const button = e.target.closest('button')
      if (!button) return
      if (!this.$keyboard.contains(button)) return
      this.dispatchKeyPressEvent(button.dataset.key)
    })

    window.addEventListener('keydown', (event) => {
      if (event.repeat === true) return
      const { key, metaKey, ctrlKey } = event
      // Crude method of ignoring things like cmd + shift + r
      if (metaKey || ctrlKey) return
      if (
        ALPHABET.includes(key.toLowerCase()) ||
        key === 'Backspace' ||
        key === 'Enter'
      ) {
        this.dispatchKeyPressEvent(key)
        // const $button = this.$keyboard.querySelector(`[data-key="${event.key}"]`);
        // $button.focus();
      }
    })

    this.$keyboard.addEventListener('transitionend', (e) => {
      // removing the transition from buttons prevents them from fading when toggling nightmode
      // console.log(e);
      const $button = e.target.closest('button')
      if (!$button) return
      if (!this.$keyboard.contains($button)) return
      $button.classList.remove('fade')
    })
    keyboardRows.forEach((row) => {
      const $keyboardRow = document.createElement('div')
      $keyboardRow.classList.add('row')
      row.forEach((key) => {
        let $el
        if (
          (key >= 'a' && key <= 'z') ||
          key === '←' ||
          key === '↵'
        ) {
          $el = keyTemplate.content.cloneNode(true).firstElementChild
          $el.dataset.key = key
          $el.textContent = key
          if (key === '←') {
            const backspaceIcon = document.createElement('game-icon')
            backspaceIcon.setAttribute('icon', 'backspace')
            $el.textContent = ''
            $el.appendChild(backspaceIcon)
            $el.classList.add('one-and-a-half')
          }
          if (key == '↵') {
            $el.textContent = 'enter'
            $el.classList.add('one-and-a-half')
          }
        } else {
          $el =
            spacerTemplate.content.cloneNode(true).firstElementChild
          $el.classList.add(key.length === 1 ? 'half' : 'one')
        }

        $keyboardRow.appendChild($el)
      })
      this.$keyboard.appendChild($keyboardRow)
    })

    this._render()
  }

  _render() {
    for (const letter in this._letterEvaluations) {
      const $button = this.$keyboard.querySelector(
        `[data-key="${letter}"]`,
      )
      $button.dataset.state = this._letterEvaluations[letter]
      $button.classList.add('fade')
    }
  }
}

customElements.define('game-keyboard', GameKeyboard)
