const gameRowTemplate = document.createElement('template')

gameRowTemplate.innerHTML = `
  <style>
    :host {
      display: block;
    }
    :host([invalid]){
      animation-name: Shake;
      animation-duration: 600ms;
    }
    .row {
      display: grid;
      grid-template-columns: repeat(5, 1fr);
      grid-gap: 5px;
    }
    .win {
      animation-name: Bounce;
      animation-duration: 1000ms;
    }

    @keyframes Bounce {
      0%, 20% {
        transform: translateY(0);
      }
      40% {
        transform: translateY(-30px);
      }
      50% {
        transform: translateY(5px);
      }
      60% {
        transform: translateY(-15px);
      }
      80% {
        transform: translateY(2px);
      }
      100% {
        transform: translateY(0);
      }
    }

    @keyframes Shake {
      10%,
      90% {
        transform: translateX(-1px);
      }

      20%,
      80% {
        transform: translateX(2px);
      }

      30%,
      50%,
      70% {
        transform: translateX(-4px);
      }

      40%,
      60% {
        transform: translateX(4px);
      }
    }
  </style>
  <div class="row"></div>
`

export class GameRow extends HTMLElement {
  constructor() {
    super()
    this.attachShadow({ mode: 'open' })
    // Should these be defined outside of the constructor?
    this._letters = ''
    this._evaluation = []
    this._length
  }

  static get observedAttributes() {
    return ['letters', 'length', 'invalid', 'win']
  }

  get evaluation() {
    return this._evaluation
  }

  set evaluation(value) {
    console.log(`${this._letters} setting evaluation property`)
    this._evaluation = value
    if (!this.$tiles) return
    this.$tiles.forEach((tile, index) => {
      tile.setAttribute('evaluation', this._evaluation[index])
      const delay = index * 300
      setTimeout(() => {
        tile.setAttribute('reveal', '')
      }, delay)
    })
  }

  connectedCallback() {
    this.shadowRoot.appendChild(
      gameRowTemplate.content.cloneNode(true),
    )
    this.$row = this.shadowRoot.querySelector('.row')
    console.groupCollapsed(`${this._letters} create tiles`)
    for (let i = 0; i < this._length; i++) {
      const $tile = document.createElement('game-tile')
      const letter = this._letters[i]
      if (letter) {
        $tile.setAttribute('letter', letter)
      }
      // If this_evaluation exists during CCB, it means we're loading from localStorage
      if (this._evaluation[i]) {
        $tile.setAttribute('evaluation', this._evaluation[i])
        // Short delay as tile evaluations in connectedCallback indicate that
        // gameState was loaded from local storage.
        const delay = i * 100
        // console.log(letter, delay);
        setTimeout(() => {
          $tile.setAttribute('reveal', '')
        }, delay)
      }
      if (i === this._length - 1) {
        $tile.last = true
      }
      this.$row.appendChild($tile)
    }

    this.$tiles = this.shadowRoot.querySelectorAll('game-tile')
    console.groupEnd()
    // console.log(this._letters + ' render from CC');
    // this._render();

    this.addEventListener('animationend', (e) => {
      if (e.animationName === 'Shake') {
        this.removeAttribute('invalid')
      }
    })
  }

  attributeChangedCallback(name, oldValue, newValue) {
    console.log('row attr changed', name, newValue)
    switch (name) {
      case 'letters': {
        this._letters = newValue || ''
        break
      }
      case 'length': {
        this._length = parseInt(newValue, 10)
        break
      }
      case 'win': {
        if (newValue === null) {
          this.$tiles.forEach(($tile) => {
            $tile.classList.remove('win')
          })
          break
        }
        this.$tiles.forEach(($tile, index) => {
          $tile.classList.add('win')
          $tile.style.animationDelay = `${index * 100}ms`
        })
        break
      }
    }
    // console.log('row render from ACC');
    this._render()
  }

  _render() {
    if (!this.$row) return
    // console.groupCollapsed('row render ' + this._letters);
    this.$tiles.forEach(($tile, index) => {
      const letter = this._letters[index]
      if (letter) {
        $tile.setAttribute('letter', letter)
      } else {
        $tile.removeAttribute('letter')
      }
    })
    // console.groupEnd();
  }
}

customElements.define('game-row', GameRow)
