import { getGameState } from './gamestate';
import { getFeedbackLink } from './utils/feedback';

const gameSettingsTemplate = document.createElement('template');

gameSettingsTemplate.innerHTML = `
  <style>
  .setting {
    display: flex;
    justify-content: space-between;
    align-items: center;
    border-bottom: 1px solid var(--color-tone-4);
    padding: 16px 0;
  }

  a, a:visited {
    color: var(--color-tone-2);
  }

  .title {
    font-size: 18px;
  }
  .text {
    padding-right: 8px;
  }
  .description {
    font-size: 12px;
    color: var(--color-tone-2);
  }

  #footnote {
    position: absolute;
    bottom: 0;
    left: 0;
    right: 0;
    padding: 16px;
    color: var(--color-tone-2);
    font-size: 12px;
    text-align: right;
    display: flex;
    justify-content: space-between;
    align-items: flex-end;
  }

  @media only screen and (min-device-width : 320px) and (max-device-width : 480px) {
    .setting {
      padding: 16px;
    }
  }

  </style>
  <div class="sections">
    <section>
      <div class="setting">
        <div class="text">
          <div class="title">Hard Mode</div>
          <div class="description">Any revealed hints must be used in subsequent guesses</div>
        </div>
        <div class="control">
          <game-switch id="hard-mode" name="hard-mode"></game-switch>
        </div>
      </div>
      <div class="setting">
        <div class="text">
          <div class="title">Dark Theme</div>
        </div>
        <div class="control">
          <game-switch id="dark-theme" name="dark-theme"></game-switch>
        </div>
      </div>
      <div class="setting">
        <div class="text">
          <div class="title">High Contrast Mode</div>
          <div class="description">For improved color vision</div>
        </div>
        <div class="control">
          <game-switch id="color-blind-theme" name="color-blind-theme"></game-switch>
        </div>
      </div>
    </section>

    <section>
      <div class="setting">
        <div class="text">
          <div class="title">Feedback</div>
        </div>
        <div class="control">
          <a href="${getFeedbackLink()}" title="nytgames@nytimes.com">Email</a>
        </div>
      </div>
      <div class="setting">
        <div class="text">
          <div class="title">Community</div>
        </div>
        <div class="control">
          <a href="https://twitter.com/NYTGames" target="blank" title="@NYTGames">Twitter</a>
        </div>
      </div>
      <div class="setting">
      <div class="text">
        <div class="title">Questions?</div>
      </div>
      <div class="control">
        <a href="https://help.nytimes.com/hc/en-us/articles/360029050872-Word-Games-and-Logic-Puzzles#h_01FVGCB2Z00ZQMDMCYWBPWJNXB" target="blank">FAQ</a>
      </div>
    </div>
    </section>
  </div>
  <div id="footnote">
    <div id="copyright">&#169; ${new Date().getFullYear()} The New York Times Company</div>
    <div>
      <div id="puzzle-number"></div>
    </div>
  </div>
`;

export class GameSettings extends HTMLElement {
  gameApp;

  constructor() {
    super();
    this.attachShadow({ mode: 'open' });
  }

  connectedCallback() {
    this.shadowRoot.appendChild(gameSettingsTemplate.content.cloneNode(true));
    this.shadowRoot.querySelector('#puzzle-number').textContent = `#${this.gameApp.dayOffset}`;

    this.shadowRoot.addEventListener('game-switch-change', (e) => {
      e.stopPropagation();
      const { name, checked, disabled } = e.detail;
      this.dispatchEvent(
        new CustomEvent('game-setting-change', {
          bubbles: true,
          composed: true,
          detail: { name, checked, disabled },
        }),
      );
      // Need to render after switching so we can disable any invalid settings
      this.render();
    });
    this.render();
  }

  render() {
    const body = document.querySelector('body');

    if (body.classList.contains('nightmode')) {
      this.shadowRoot.querySelector('#dark-theme').setAttribute('checked', '');
    }
    if (body.classList.contains('colorblind')) {
      this.shadowRoot.querySelector('#color-blind-theme').setAttribute('checked', '');
    }
    const gameState = getGameState();
    if (gameState.hardMode) {
      this.shadowRoot.querySelector('#hard-mode').setAttribute('checked', '');
    }
    // This bullshit is because we need to disable switching on hard mode
    // mid-round but we don't have a good flow for state, so have to put a bunch
    // of game related logic down here
    if (!gameState.hardMode && gameState.gameStatus === 'IN_PROGRESS' && gameState.rowIndex !== 0) {
      this.shadowRoot.querySelector('#hard-mode').removeAttribute('checked');
      this.shadowRoot.querySelector('#hard-mode').setAttribute('disabled', '');
    }
  }
}

customElements.define('game-settings', GameSettings);
