const template = document.createElement('template')
template.innerHTML = `
<style>
  :host {
    display: inline-block;
  }
  .tile {
    width: 100%;
    display: inline-flex;
    justify-content: center;
    align-items: center;
    font-size: 2rem;
    line-height: 2rem;
    font-weight: bold;
    vertical-align: middle;
    box-sizing: border-box;
    color: var(--tile-text-color);
    text-transform: uppercase;
    user-select: none;
  }
  .tile::before {
    content: '';
    display: inline-block;
    padding-bottom: 100%;
  }

  /* Allow tiles to be smaller on small screens */
  @media (max-height: 600px) {
    .tile {
      font-size: 1em;
      line-height: 1em;
    }
  }

  .tile[data-state='empty'] {
    border: 2px solid var(--color-tone-4);
  }
  .tile[data-state='tbd'] {
    background-color: var(--color-tone-7);
    border: 2px solid var(--color-tone-3);
    color: var(--color-tone-1);
  }
  .tile[data-state='correct'] {
    background-color: var(--color-correct);
  }
  .tile[data-state='present'] {
    background-color: var(--color-present);
  }
  .tile[data-state='absent'] {
    background-color: var(--color-absent);
  }

  .tile[data-animation='pop'] {
    animation-name: PopIn;
    animation-duration: 100ms;
  }

  @keyframes PopIn {
    from {
      transform: scale(0.8);
      opacity: 0;
    }

    40% {
      transform: scale(1.1);
      opacity: 1;
    }
  }
  .tile[data-animation='flip-in'] {
    animation-name: FlipIn;
    animation-duration: 250ms;
    animation-timing-function: ease-in;
  }
  @keyframes FlipIn {
    0% {
      transform: rotateX(0);
    }
    100% {
      transform: rotateX(-90deg);
    }
  }
  .tile[data-animation='flip-out'] {
    animation-name: FlipOut;
    animation-duration: 250ms;
    animation-timing-function: ease-in;
  }
  @keyframes FlipOut {
    0% {
      transform: rotateX(-90deg);
    }
    100% {
      transform: rotateX(0);
    }
  }
</style>
<div class="tile" data-state="empty" data-animation="idle"></div>
`

export class GameTile extends HTMLElement {
  _letter = ''
  _state = 'empty'
  _animation = 'idle'
  _last = false
  _reveal = false

  constructor() {
    super()
    this.attachShadow({ mode: 'open' })
  }

  static get observedAttributes() {
    return ['letter', 'evaluation', 'reveal']
  }

  set last(value) {
    // console.log('setting last');
    this._last = value
  }

  connectedCallback() {
    // console.log('connected callback');
    this.shadowRoot.appendChild(template.content.cloneNode(true))
    this.$tile = this.shadowRoot.querySelector('.tile')

    this.$tile.addEventListener('animationend', (e) => {
      // console.log(this._letter, e.animationName);
      if (e.animationName === 'PopIn') {
        this._animation = 'idle'
      }
      if (e.animationName === 'FlipIn') {
        this.$tile.dataset.state = this._state
        this._animation = 'flip-out'
      }
      if (e.animationName === 'FlipOut') {
        this._animation = 'idle'
        if (this._last) {
          this.dispatchEvent(
            new CustomEvent('game-last-tile-revealed-in-row', {
              bubbles: true,
              composed: true,
            }),
          )
        }
      }
      // console.log(`tile ${this._letter} render from animationEnd cc`);
      this._render()
    })
    // console.log(`tile ${this._letter} render from cc`);
    this._render()
  }

  attributeChangedCallback(name, oldValue, newValue) {
    // console.log('tile attr changed', name, oldValue, newValue);
    switch (name) {
      case 'letter': {
        // Don't do anything if letter value hasn't changed
        if (newValue === oldValue) break
        const letter = newValue === 'null' ? '' : newValue
        this._letter = letter
        this._state = letter ? 'tbd' : 'empty'
        this._animation = letter ? 'pop' : 'idle'
        break
      }
      case 'evaluation': {
        if (!newValue) break
        this._state = newValue
        break
      }
      case 'reveal': {
        this._animation = 'flip-in'
        this._reveal = true
        break
      }
    }
    // console.log(`tile ${this._letter} render from acc`);
    this._render()
  }

  _render() {
    if (!this.$tile) {
      // console.log('elem doesnt exist, aborting render');
      return
    }

    this.$tile.textContent = this._letter

    if (['empty', 'tbd'].includes(this._state)) {
      this.$tile.dataset.state = this._state
    }

    if (['empty', 'tbd'].includes(this._state) || this._reveal) {
      if (this.$tile.dataset.animation != this._animation) {
        this.$tile.dataset.animation = this._animation
      }
    }
  }
}

customElements.define('game-tile', GameTile)
